@@ -65,13 +65,19 @@ int verify_pbkdf_params(struct crypt_device *cd,
 	const char *pbkdf_type;
 	int r = 0;
 
-	if (!pbkdf->type || !pbkdf->hash || !pbkdf->time_ms)
+	if (!pbkdf->type ||
+	    (!pbkdf->hash && !strcmp(pbkdf->type, "pbkdf2")))
 		return -EINVAL;
 
+	if (!pbkdf->time_ms && !(pbkdf->flags & CRYPT_PBKDF_NO_BENCHMARK)) {
+		log_err(cd, _("Requested PBKDF target time cannot be zero."));
+		return -EINVAL;
+	}
+
 	/* TODO: initialise crypto and check the hash and pbkdf are both available */
 	r = crypt_parse_pbkdf(pbkdf->type, &pbkdf_type);
 	if (r < 0) {
-		log_err(cd, _("Unknown PBKDF type %s.\n"), pbkdf->type);
+		log_err(cd, _("Unknown PBKDF type %s."), pbkdf->type);
 		return r;
 	}
 
@@ -82,18 +88,18 @@ int verify_pbkdf_params(struct crypt_device *cd,
 	if (crypt_get_type(cd) &&
 	    !strcmp(crypt_get_type(cd), CRYPT_LUKS1) &&
 	    strcmp(pbkdf_type, CRYPT_KDF_PBKDF2)) {
-		log_err(cd, _("Requested PBKDF type is not supported for LUKS1.\n"));
+		log_err(cd, _("Requested PBKDF type is not supported for LUKS1."));
 		return -EINVAL;
 	}
 
 	if (!strcmp(pbkdf_type, CRYPT_KDF_PBKDF2)) {
 		if (pbkdf->max_memory_kb || pbkdf->parallel_threads) {
-			log_err(cd, _("PBKDF max memory or parallel threads must not be set with pbkdf2.\n"));
+			log_err(cd, _("PBKDF max memory or parallel threads must not be set with pbkdf2."));
 			return -EINVAL;
 		}
 		if (pbkdf->flags & CRYPT_PBKDF_NO_BENCHMARK &&
 		    pbkdf->iterations < pbkdf_limits.min_iterations) {
-			log_err(cd, _("Forced iteration count is too low for %s (minimum is %u).\n"),
+			log_err(cd, _("Forced iteration count is too low for %s (minimum is %u)."),
 				pbkdf_type, pbkdf_limits.min_iterations);
 			return -EINVAL;
 		}
@@ -103,32 +109,28 @@ int verify_pbkdf_params(struct crypt_device *cd,
 	/* TODO: properly define minimal iterations and also minimal memory values */
 	if (pbkdf->flags & CRYPT_PBKDF_NO_BENCHMARK) {
 		if (pbkdf->iterations < pbkdf_limits.min_iterations) {
-			log_err(cd, _("Forced iteration count is too low for %s (minimum is %u).\n"),
+			log_err(cd, _("Forced iteration count is too low for %s (minimum is %u)."),
 				pbkdf_type, pbkdf_limits.min_iterations);
 			r = -EINVAL;
 		}
 		if (pbkdf->max_memory_kb < pbkdf_limits.min_memory) {
-			log_err(cd, _("Forced memory cost is too low for %s (minimum is %u kilobytes).\n"),
+			log_err(cd, _("Forced memory cost is too low for %s (minimum is %u kilobytes)."),
 				pbkdf_type, pbkdf_limits.min_memory);
 			r = -EINVAL;
 		}
 	}
 
 	if (pbkdf->max_memory_kb > pbkdf_limits.max_memory) {
-		log_err(cd, _("Requested maximum PBKDF memory cost is too high (maximum is %d kilobytes).\n"),
+		log_err(cd, _("Requested maximum PBKDF memory cost is too high (maximum is %d kilobytes)."),
 			pbkdf_limits.max_memory);
 		r = -EINVAL;
 	}
 	if (!pbkdf->max_memory_kb) {
-		log_err(cd, _("Requested maximum PBKDF memory can not be zero.\n"));
+		log_err(cd, _("Requested maximum PBKDF memory cannot be zero."));
 		r = -EINVAL;
 	}
 	if (!pbkdf->parallel_threads) {
-		log_err(cd, _("Requested PBKDF parallel threads can not be zero.\n"));
-		r = -EINVAL;
-	}
-	if (!pbkdf->time_ms) {
-		log_err(cd, _("Requested PBKDF target time can not be zero.\n"));
+		log_err(cd, _("Requested PBKDF parallel threads cannot be zero."));
 		r = -EINVAL;
 	}
 
@@ -165,9 +167,9 @@ int init_pbkdf_type(struct crypt_device *cd,
 	 * It will fail later anyway :-)
 	 */
 	type = strdup(pbkdf->type);
-	hash = strdup(pbkdf->hash);
+	hash = pbkdf->hash ? strdup(pbkdf->hash) : NULL;
 
-	if (!type || !hash) {
+	if (!type || (!hash && pbkdf->hash)) {
 		free(CONST_CAST(void*)type);
 		free(CONST_CAST(void*)hash);
 		return -ENOMEM;
@@ -251,6 +253,19 @@ const struct crypt_pbkdf_type *crypt_get_pbkdf_type(struct crypt_device *cd)
 	return crypt_get_pbkdf(cd)->type ? crypt_get_pbkdf(cd) : NULL;
 }
 
+const struct crypt_pbkdf_type *crypt_get_pbkdf_default(const char *type)
+{
+	if (!type)
+		return NULL;
+
+	if (!strcmp(type, CRYPT_LUKS1))
+		return &default_luks1;
+	else if (!strcmp(type, CRYPT_LUKS2))
+		return &default_luks2;
+
+	return NULL;
+}
+
 void crypt_set_iteration_time(struct crypt_device *cd, uint64_t iteration_time_ms)
 {
 	struct crypt_pbkdf_type *pbkdf;
